/*
 * ChordNodeImp.java
 * 
 * @author  Paul Jackson
 */
package dhash;

import java.util.logging.*;

import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;

// Since all files in local filesystem, shouldn't need this.
// import java.rmi.RMISecurityManager;

	

/**
 * Single-threaded implementation of Chord protocol.
 *
 * Based closely on presentation of non-concurrent join algorithm in:
 *
 *    Chord: a scaleable peer-to-peer lookup service for internet applications.
 *    I. Stoica, R. Morris, D. Karger, M.F. Kaashoek, H. Balakrishnan.
 *    ACM SigComm 2001. Aug 27-31 2001.
 *
 * A very similar presentation is in MIT Tech Report 819, same title and
 * authors, 2001.
 *
 * Variations:
 *
 * 1. Changes in notation:
 *
 *    Paper              Here.
 *    finger[i].node     finger[i] 
 *    finger[i].start    nodeId.addPowerTwo(i-1)
 *    x \in (y,z)        x.inOpenOpenRange(y,z)
 *    x \in (y,z]        x.inOpenClosedRange(y,z)
 *
 * 2. The paper is not consistent on whether the terms `predecessor' and 
 * `successor' are strict or not.  This is rather confusing. This 
 * implementation adopts the conventions that:
 *   succ/pred of a Node is always strict.
 *   succ/pred of an ID is non strict by default
 *   strictSucc strictPred used when want to refer to strict succ/pred of an ID
 *
 * 3. See the file ID.java for more on the definition of predicates such
 * as inOpenOpenRange() and resolution of ambiguities.
 *
 */

public class ChordNodeImp extends UnicastRemoteObject implements ChordNode {

    //== Static fields ======================================================= 
    
    // See Serializable API docs. UnicastRemoteObject implements it.
    private static final long serialVersionUID = 42L;

    private static Logger logger = Logger.getLogger("dhash.ChordNodeImp");
    

    private static int next = 0;     // Counter to help with debugging.
    
    //== Instance fields ======================================================
    private ID nodeId;     // Node identifier
    private String name;   // Readable name, unique to Java process.
    private String domain; // Readable name for process, globally unique.
                           // In distributed setting, could be URL + process ID
    
    private ChordNodeRef predecessor;
    
    private ChordNodeRef[] finger; // Indexed from 1 .. ID.getSize(). 
                                   // [0] unused.
    
    // appNode = application node.  Used by ChordNodeImp object to signal to
    // associated application node that it has a new predecessor.
    
    private AppNode appNode;
    
    // Object numbering to help with debugging.
    public int objNum = next++;
    
    //== Constructors ========================================================
    /** Create a new instance of ChordNodeImp
 
     The name and domain of a ChordNodeImp object are expected to be constant.
     If they are changed, they will get out of sync with ChordNodeRefs 
     Referring to the object.
     */

    public ChordNodeImp(String name, String domain, ID id)
             throws RemoteException
    {
        super(); // Export object so available to receive RMI calls.
                 // javac supplies call if not explicit.  Made explicit
                 // here for clarity.
        this.finger = new ChordNodeRef[ID.getSize() + 1];
        this.name = name;
        this.domain = domain;
        this.nodeId = id;
    }
    
    //== Methods =============================================================

    //- Convenience wrappers for logging calls ----------------------------

    private void log(String methodName, String message, Object[] params) {
        logger.logp(Level.FINE, "dhash.ChordNodeImp",
                    methodName, message, params);
    }
    private void log(String methodName, String msg, Object p1) {
        log(methodName, msg, new Object[]{p1});
    }
    private void log(String methodName, String msg, Object p1, Object p2) {
        log(methodName, msg, new Object[]{p1,p2});
    }
    private void log(String methodName, String msg,
                     Object p1, Object p2, Object p3) {
        log(methodName, msg, new Object[]{p1,p2,p3});
    }

    //-- Setters and getters ------------------------------------------------

    public ID getID() {return nodeId;}
    
    public String getName() {return name;}
    public String getDomain() {return domain;}
    
    public ChordNodeRef getSuccessor() {return finger[1];}
    public ChordNodeRef getPredecessor() {return predecessor;}
    public void setPredecessor(ChordNodeRef n) {predecessor = n;}
    
    public ChordNodeRef getFinger(int i) {return finger[i];}


    //-- Primary methods implementing Chord protocol -------------------------

    /** Ask this node to find successor of id */
    
    public ChordNodeRef findSuccessor(ID id) throws RemoteException {

        log("findSuccessor", "ENTR", this, id);
        log("findSuccessor", "CALL", "findStrictPredecessor", this, id);
                        
        ChordNodeRef nprime = findStrictPredecessor(id);

        log("findSuccessor", "RCV", nprime);

        ChordNodeRef nprimeSucc = nprime.getSuccessor();
        
        log("findSuccessor", "RTN", nprimeSucc);
        return nprimeSucc;
    }
    
    /** Ask this node to find strict predecessor of id

      This procedure is iterative in the orginal Chord paper. This
      recursive version is suggested in the 2003 IEEE Transactions on
      Networking journal paper on Chord (draft available from the
      Chord website).  Its advantage is that it roughly halves the
      number of remote calls required by a search.

    */
    
    public ChordNodeRef findStrictPredecessor(ID id) throws RemoteException {
        log("findStrictPredecessor", "ENTR", this, id);
        
        if (id.inOpenClosedRange(nodeId, getSuccessor().getID() )) {

            log("findStrictPredecessor", "RTN", this);

            return new ChordNodeRef(this);

        } else {
            log("findStrictPredecessor", "CALL",
                "closestStrictlyPrecedingFinger", this, id);

            ChordNodeRef nprime = closestStrictlyPrecedingFinger(id);
            
            log("findStrictPredecessor", "RCV", nprime);
            log("findStrictPredecessor", "CALL",
                "findStrictPredecessor", nprime, id);

            ChordNodeRef result = nprime.findStrictPredecessor(id);

            log("findStrictPredecessor", "RCV", result);
            log("findStrictPredecessor", "RTN", result);

            return result;
        }
    }
    /**
       closestStrictlyPrecedingFinger()

       Notes on operation: 

       -  Instead of loop termination test  `id in (finger[i], nodeId]*'
          paper uses  `finger[i] in (nodeId, id)'. These tests are exactly
          logically equivalent, but correctness is easier to see for the
          test supplied here.

       -  Why is there no danger of reaching the warning at the method end?

          Note that the method is only called when nodeId is not
          responsible for id, i.e. when `! id in (nodeId, finger[1]]*'.
          This condition is equivalent to `id in (finger[1], nodeId]*', which
          is the loop termination test for the last loop iteration.  Hence
          the loop will terminate by the last iteration at the latest, and
          never normally exits.

       -  When an iteration terminates, we know that the termination test
          was false on previous iterations.  Knowing this, and that fingers 
          are ordered around the circle of IDs, we can see that the 
          termination test when i < m is equivalent to
          `id in (finger[i], finger[i+1] ]*'.  Hence we can see that the 
          node returned really is the closest finger strictly preceding id.

    */
    
    private ChordNodeRef closestStrictlyPrecedingFinger(ID id) {

        log("closestStrictlyPrecedingFinger", "ENTR", this, id);

        final int m = ID.getSize();

        for (int i = m; i > 0; i--) {
            if (  id.inOpenClosedStarRange(finger[i].getID(), nodeId) ) {

                log("closestStrictlyPrecedingFinger", "INFO",
                    "stopped at finger", new Integer(i));
                log("closestStrictlyPrecedingFinger", "RTN", finger[i]);
            
                return finger[i];
            }
        }

        logger.warning("Failed to find any strictly preceding finger");
        return null; 
        
    }
    
    public ChordNodeRef findPredecessor(ID id) throws RemoteException {
        ID idPlusOne = id.addOne();

        log("findPredecessor", "ENTR", this, id);
        log("findPredecessor", "CALL",
            "findStrictPredecessor", this, idPlusOne);
        
        ChordNodeRef result = findStrictPredecessor(id.addOne());

        log("findPredecessor", "RCV", result);
        log("findPredecessor", "RTN", result);

        return result;
    }
    
    
    /** Initialize newly created node for new ring.  */
    
    public void initForNewRing() {
        predecessor = new ChordNodeRef(this);
        for (int i = 1; i < finger.length; i++) finger[i] = predecessor;
    }
    
    /**
     * Join a newly created node to an existing ring.
     *
     * Just do case in paper when n != null.  
     * Null case code is in previous method.
     */
    
    public void join(ChordNodeRef nprime) throws RemoteException {
        logger.entering("dhash.ChordNodeImp", "join",
                         new Object[]{this, nprime} );
        
        // findSuccessor call here is in initFingerTable in paper.
        // Paper finds the strict successor.  Here we don't so that we can 
        // detect clashes.
        
        ChordNodeRef succ = nprime.findSuccessor(nodeId);
        if (nodeId == succ.getID()) {
            throw new IllegalArgumentException(
              "Trying to join ring with new nodes' ID already in use");
        }
        initFingerTable(nprime, succ);
        updateOthers();
        
        // Signal to successor that predecessor is new.
        // If key/value pairs are being stored on 
        // successor, some of them will need moving to
        // new predecessor. 
        succ.processNewPredecessor(); 
        
    }
    
    /** 
     * Initialise finger table and successor and predecessor pointers.
     *
     * Initial retrieval of successor moved out of this method into join
     * method for better exception handling.
     *
     * @param nprime  some node already in ring.
     * @param succ    successor in ring of node just being added.
     *
     * finger[i] set to successor of start[i] in ring *before* new node
     * inserted: findSuccessor() never sees new node, because it never
     * notices modified predecessor pointer on successor of new node.
     *
     * updateOthers() will fix fingers to account for new node.
     */
    
    public void initFingerTable(ChordNodeRef nprime, ChordNodeRef succ) 
            throws RemoteException {
        logger.entering("dhash.ChordNodeImp", "join",
                         new Object[]{this, nprime, succ} );
        
        finger[1] = succ; // Set successor
        predecessor = succ.getPredecessor();
        succ.setPredecessor(new ChordNodeRef(this));
        
        for (int i = 2; i < finger.length; i++) {
            // All fingers with starts before or at this.successor
            // get initialised the same.
            if (nodeId.addPowerTwo(i-1).inOpenClosedRange(nodeId, 
                                                  finger[i-1].getID())
                ) {
                finger[i] = finger[i-1];                                
            } else {
                // Find successor in old ring of this.finger[i].start

                ID fingerIStart = nodeId.addPowerTwo(i-1);
                finger[i] = nprime.findSuccessor(fingerIStart);
            }
        }
    
    }
    /** 
     * Update all nodes whose finger tables might need to refer to this.
     * This possibly *includes* the new node.
     *
     * Called notify() in TR.
     * 
     * At first sight, it is slightly alarming that ring search using the
     * finger tables is used *while* fingers are being updated.
     * 
     * However, the updates are benign: one property they enjoy is that
     * they always keep the fingers in finger tables in their proper order.
     * 
     * Also, the search algorithm is tolerant of fingers not always pointing
     * to the first node after their start position: if fingers are slightly
     * out of place, convergence might be slightly slower, but correct
     * termination of the search is still guaranteed.
     */
    
    public void updateOthers() throws RemoteException {
        // System.out.println("updateOthers(): call on " + getID());

        logger.entering("dhash.ChordNodeImp", "updateOthers");  
        
        for (int i = 1; i < finger.length; i++) {

            // NB: findPredecessor in paper is strict.
            // Calling it is a bug. Here findPredecessor is non-strict
            // which is good.
            ChordNodeRef p = findPredecessor(nodeId.subtractPowerTwo(i-1));
            p.updateFingerTable(new ChordNodeRef(this),i);
            
        }
    }
    /**
       Possibly update this.finger[i] to account for new node s in ring.

       If new node s is between this.finger[i].start and current
       this.finger[i], this.finger[i] is set to s.


    */
    
    public void updateFingerTable(ChordNodeRef s, int i)
        throws RemoteException
    {

        logger.entering("dhash.ChordNodeImp", "updateFingerTable",
                         new Object[]{this, s, new Integer(i)} );


        ID startId = nodeId.addPowerTwo(i - 1);

        // Paper uses nodeId in place of startId, but correctness is then less
        // transparent in case when updateFingerTable is called on new node.
        // 
        if (s.getID().inClosedOpenStarRange(startId, finger[i].getID() )) {

            finger[i] = s;
            ChordNodeRef p = predecessor; 
            p.updateFingerTable(s, i);
        }
        
    }
   
    //-- Printing nodes -------------------------------------------------------
    
    // Printing policy:
    // 1. default toString() gives
    //    a) info on RMI server for node.
    //    b) local info
    
    // 2. RMI accessible toShortString() gives same info as toString()
    
    // 3. RMI accessible toLongString() gives same info as toString() + 
    //    info on predecessor and finger table entries.
    //    Each of these could be local or remote.
    //    In either case the toShortString() info is returned.
    //    In the proxy case, we want to see info on proxy as well.
    //      i.e. use the toString on the proxy.
    //
    
    private static String nl = System.getProperty("line.separator");
    
    public String toString() {
        String s = super.toString();
        s += "[ "  + objNum + ": " + name + ", " + nodeId + " ]";
        return s;
    }
    /**
     * Version of toString() for inclusion in ChordNode interface.  
     * 
     * toString() cannot be included in ChordNode interface because it
     * does not throw a RemoteException.
     *
     */ 
    
    public String toShortString() {return toString();}
    
    /**
     * Utility method for printing toShortString() info on node referred to,
     * along with info on proxy if node is not immediate ChordNodeImp but is
     * proxy.
     *
     *
     */
    private static String getShortString(ChordNodeRef cN) 
            throws RemoteException {

        String s = cN.toShortString();
        if (cN.isProxy() ) {
            s += "Proxy:" + nl;
            s += cN.toString();
        }
        return s;
    }
    
    public String toLongString() throws RemoteException {
        
        String s = toString();
        
        if (getSuccessor() != null) {
            s+= "successor = " + nl + getShortString(getSuccessor());
        }
        else s+= "successor = null" + nl;
        
        if (predecessor != null) {
            s+= "predecessor = " + nl + getShortString(predecessor);
        }
        else s+= "predecessor = null" + nl;
        
        s+= "fingers:" + nl;
        for (int i = 1; i < finger.length; i++) {
            if (finger[i] != null)
                 s+= "finger " + i + ":" + nl + getShortString(finger[i]); 
            else
                s+= "finger  " + i + ": null" + nl;
        }
        return s;
    }
    
    //-- Handling new predecessors -------------------------------------------
    
    public void setAppNode(AppNode n) { appNode = n; }
    public AppNode getAppNode() {return appNode; }
    
    public void processNewPredecessor() throws RemoteException {
        if (appNode != null) {
            appNode.processNewPredecessor();
        }
    }
    
    
        
}
