/*
 * ChordNodeRef.java
 *
 * @author  Paul Jackson
 */

package dhash;

import java.io.Serializable;
import java.rmi.RemoteException;

import util.AuxProcess;

/**
 *
 * A ChordNodeRef (Chord node reference) object wraps a ChordNode
 * reference, caching constant field values to ensure fast access to
 * these values without inter-process traffic.
 *
 * ChordNodeRefs implement all the methods in the ChordNode interface, but
 * importantly we do not record in the class definition that
 * 
 * ChordNodeRef implements ChordNode.
 * 
 * If we did this, Java would pass round remote references to ChordNodeRefs
 * in a multi-process setting, which is not what we want: we want these
 * ChordNodeRef objects to be serialised along with their cached values.
 * 
 * Note that, when a ChordNodeRef wrapping a ChordNodeImp object is
 * passed to a remote process, serialisation will update the chordNode
 * field to point to a proxy object that contains a remote object
 * reference to the ChordNodeImp object.
 * 
 */

public class ChordNodeRef implements Serializable {

    //== Static fields ========================================================

    // See Serializable API docs. UnicastRemoteObject implements it.
    private static final long serialVersionUID = 42L;

    //== Instance fields ======================================================
    
    private ChordNode chordNode;
    private ID nodeId;
    private String name;
    private String domain;

    //== Constructors =========================================================

    public ChordNodeRef(ChordNodeImp cMI) {
        chordNode = cMI;
        nodeId = cMI.getID();
        name = cMI.getName();
        domain = cMI.getDomain();
    }

    //== Instance Methods =====================================================

    // Check kind of reference

    // isProxy() and isRemote() do not necessarily compute the same values:
    // sometimes JavaRMI uses proxies for non-remote objects.

    public boolean isProxy() {
        return ! (chordNode instanceof ChordNodeImp);
    }
    public boolean isRemote() {
        return ! domain.equals(AuxProcess.getJVMName());
    }

    // Duplicates of methods declared in the ChordNode interface

    public ID getID() {
        return nodeId;
    }
    public String getName() {
        return name;
    }
    public String getDomain() {
        return domain;
    }
    public ChordNodeRef getSuccessor() throws RemoteException {
        return chordNode.getSuccessor();
    }
    public ChordNodeRef getPredecessor() throws RemoteException {
        return chordNode.getPredecessor();
    }
    public ChordNodeRef getFinger(int i) throws RemoteException {
        return chordNode.getFinger(i);
    }
    public void setPredecessor(ChordNodeRef n) throws RemoteException {
        chordNode.setPredecessor(n);
        return;
    }
    public ChordNodeRef findSuccessor(ID id) throws RemoteException {
        return chordNode.findSuccessor(id);
    }
    public ChordNodeRef findStrictPredecessor(ID id) throws RemoteException {
        return chordNode.findStrictPredecessor(id);
    }
    public void updateFingerTable(ChordNodeRef s, int i)
            throws RemoteException  {
        
        chordNode.updateFingerTable(s,i);
        return;
    }
    public String toShortString() throws RemoteException {
        return chordNode.toShortString();
    }
    public String toLongString() throws RemoteException {
        return chordNode.toLongString();
    }
    public AppNode getAppNode() throws RemoteException {
        return chordNode.getAppNode();
    }
    public void processNewPredecessor() throws RemoteException {
        chordNode.processNewPredecessor();
    }


    // Utility
    
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append( "ChordNodeRef [ nodeId: " + nodeId);
        sb.append( ", name: " + name);
        sb.append( ", chordNode: " + chordNode.toString());
        sb.append( " ]");
        return sb.toString();
    }


}

