/*
 * DHashNodeImp.java
 *
 *
 * @author  Paul Jackson
 *
 * Implementation for distributed hash table, built on top of the Chord
 * ring lookup service.
 *
 *
 */

package dhash;

import java.util.*;

import util.Util; // for SHA-1 digest methods.

import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;


public class DHashNodeImp extends UnicastRemoteObject 
        implements DHashNode, DHashNodeClient, AppNode {
            
    //== Static fields / initializers =========================================
    
    // See Serializable API docs. UnicastRemoteObject implements it.
    private static final long serialVersionUID = 42L;

    //== Instance fields ======================================================
    
    private HashMap<String,Object> map; // For key, value pairs
    
    private ChordNodeImp chordNodeImp; // The associated ChordNode object.
    
    private DHAppNode dHAppNode;  // dHAppNode = DHash application node. 

    //== Constructors =========================================================
    
    /** Create a new instance of DHashNodeImp 
     *
     * For ID uses digest of name, as necessary truncated to ID size.
     *
     */
    public DHashNodeImp(String name, String domain) throws RemoteException {
        this(name, domain, new ID(Util.digest(domain + ":" + name)) );
        
    }
    
    public DHashNodeImp(String name, String domain, ID id)
        throws RemoteException
    {
        super(); // Export object so available to receive RMI calls.
                 // javac supplies call if not explicit.  Made explicit
                 // here for clarity.
        
        chordNodeImp = new ChordNodeImp(name, domain, id);
        
        chordNodeImp.setAppNode(this);// Register this for callbacks
        
        map = new HashMap<String,Object>();
        
    }
    //== Methods ==============================================================
    
    //-- Control methods ----------------------------------------------------
    // (non static methods included in DHashNodeClient interface)
    // Used by external clients
    
    public static void setIDSize(int i) { 

        ID.setSize(i);
    }
    
    public void initForNewRing() {

        chordNodeImp.initForNewRing();
    }
    
    public void join(DHashNode dHN) throws RemoteException {
                                    
        chordNodeImp.join(dHN.getChordNode());
    }

    //-- ChordNode Access methods  -------------------------------------------
    // (included in DHashNode interface)
    
    public ChordNodeRef getChordNode() {return new ChordNodeRef(chordNodeImp);}
    
    // Used in logging messages
    public String getName() {return chordNodeImp.getName();}
    public String getDomain() {return chordNodeImp.getDomain();}
    
    //-- LocalMap methods ---------------------------------------------------
    // (included in DHashNode interface)
    
    // Used by this and other DHashNodeImps to access a node's local map.
    
    public void localPut(String key, Object value) {
        
        map.put(key, value);
    }

    public Object localGet(String key) {          // returns null if no entry.

        return map.get(key);
    }

    public void localDHAppGet(String key) throws RemoteException {   

        Object result = map.get(key);
        dHAppNode.processValue(result);
        return;
    }

    public boolean localContains(String key) {

        return map.containsKey(key);
    }

    public void localRemove(String key)  {
        
        map.remove(key);
    }

    // For debugging.

    public Map<String,Object> getLocalMap() {return map;}

    //-- DHashMap interface ---------------------------------------------------
    // (included in DHashNodeClient interface)
    // Used by external clients
    
    /** Find the DHashNode associated with key */
    
    private DHashNode findDHashNode(String key) throws RemoteException {

        ID id = new ID(Util.digest(key));
        ChordNodeRef cNodeWithKey = chordNodeImp.findSuccessor(id);
        
        DHashNode dHNodeWithKey = (DHashNode) cNodeWithKey.getAppNode();
        
        return dHNodeWithKey;
    }
    
    public void put(String key, Object value) throws RemoteException {

        DHashNode dHN = findDHashNode(key);
        dHN.localPut(key, value);
    }
    
    // returns null if no entry.
    public Object get(String key) throws RemoteException { 

        DHashNode dHN = findDHashNode(key);
        return dHN.localGet(key);
    }
    
    public boolean contains(String key) throws RemoteException {

        DHashNode dHN = findDHashNode(key);
        return dHN.localContains(key);
    }

    public void remove(String key) throws RemoteException {

        DHashNode dHN = findDHashNode(key);
        dHN.localRemove(key);
    }

    // Transfer found value to DHAppNode associated with local map holding
    // key value pair.  and return reference to DHAppNode.
    
    public DHAppNode dHAppGet(String key) throws RemoteException { 
        DHashNode dHN = findDHashNode(key);
        dHN.localDHAppGet(key);
        
        return dHN.getDHAppNode();
        
    }
    





    // -- CallBack interface --------------------------------------------------
 
    // Invoked by chordNodeImp when it has a new predecessor.
    // Move key/value pairs that belong on new predecessor to it
    
    public void processNewPredecessor() throws RemoteException {
        ChordNodeRef newPred = chordNodeImp.getPredecessor();
        ID newPredID = newPred.getID();
        ID thisID = chordNodeImp.getID();
        DHashNode predDHN = (DHashNode) newPred.getAppNode();
                
        for (Iterator i = map.entrySet().iterator(); i.hasNext(); ) {
            Map.Entry e = (Map.Entry) i.next();
            String eKey = (String) e.getKey();
            Object eValue = e.getValue();
          
            // Could be more efficient instead to store IDs paired
            // with values in map HashMap.  But new nodes join
            // infrequently and code here is simpler.
            
            ID eKeyID = new ID(Util.digest(eKey));
        
            // if id for e.key <= predecessor ID, then move it to 
            // predecessor
          
            if (! eKeyID.inOpenClosedRange(newPredID, thisID)) {
                i.remove();
                
                predDHN.localPut(eKey, eValue);
            }
        }
    }

    //-- Interfacing with DHAppNode -----------------------------------------
    
    public void setDHAppNode(DHAppNode n) { dHAppNode = n; }
    public DHAppNode getDHAppNode() {return dHAppNode; }

    
}
