/*
 * ID.java
 * 
 * @author  Paul Jackson
 */

package dhash;

import java.math.BigInteger;
import java.util.logging.*;

import java.io.Serializable;
/**
 * Class for key and node identifiers in a Chord ring
 *
 * We consider ID values to increase in moving clockwise around the ring.
 *
 * Since IDs are values on a circle, it does not make sense to ask whether
 * one ID is `less than' or before another ID.
 * 
 * Instead we need to ask 3-argument questions about whether or not
 * one ID is in the interval of IDs defined by 2 other IDs.
 * Alternative definitions of intervals are possible, depending on
 * whether we include each of the IDs at each end in the interval.
 *
 * Using terminology from mathematics, we say that an interval limit
 * is `closed' if the limit is included in the range and `open' if the
 * limit is not.  When writing intervals, we use a
 *
 *   `[' for a closed interval start,
 *   `(' for an open interval start,
 *   `]' for a closed interval end,
 *   `)' for an open interval end.
 *
 * In using interval notation on a ring, we treat a lower open limit "(i"
 * (a lower limit excluding i) as a closed lower limit "[i + delta", 
 * (a lower limit including i + delta) where 0 < delta << 1.
 * Similarly, we treat a higher open limit "i)"
 * as a closed lower limit "i - delta]".
 * 
 * This convention allows us to easily say what values are in an interval
 * when its start and end are equal. For example, we have:
 *               
 * (0,2)  is {1}
 * (0,1)  is {}  
 * (0,0)  is S - {0}  
 * (0,-1) is S - {-1,0}
 * 
 * [0,1)  is {0}
 * [0,0)  is S 
 * [0,-1) is S - {-1}
 * 
 * (0,1]  is {1}
 * (0,0]  is S 
 * (0,-1] is S - {0}
 * 
 * where S is the set of all ID values.
 * 
 * It is useful to define variations on the (i,j] and [i,j) ranges
 * that are empty rather than the full set S when i = j.  We choose to
 * call these intervals `star' intervals (suggestions for a better name
 * are welcome) and use the notation (i,j]* and [i,j)*.
 * 
 * 
 *
 */

public class ID implements Serializable {
    
    //== Static fields ========================================================

    // See Serializable API docs. UnicastRemoteObject implements it.
    private static final long serialVersionUID = 42L;

    private static Logger logger = Logger.getLogger("dhash.ID");
   
    /** Number of bits of id */
    private static int size = 160; // Size of SHA-1 digests
    
    /** Upper bound on ID value */
    private static BigInteger modulus = powerTwo(size);
    
    //== Instance fields ======================================================
    
    /** Numerical value of ID */
    private BigInteger val;
    
    //== Constructors =========================================================
    /** Creates a new instance of ID */
    public ID(BigInteger i) { val = i.mod(modulus); }
    
    public ID(byte[] bytes) { this( new BigInteger(bytes)); }
    
    
    /** Methods useful for explicit construction of IDs */
    public ID(String s){ this( new BigInteger(s)); }
    public ID(String s, int radix) { this( new BigInteger(s, radix));}
    
    //== Methods ==============================================================
    
    private static BigInteger powerTwo(int i) {
        return BigInteger.ONE.shiftLeft(i);
    }
    /** 
     * Set number of bits for all IDs 
     */
    public static void setSize(int size) {
        ID.size = size;
        ID.modulus = powerTwo(size);
    }
    
    public static int getSize() {return size;}
    
    
    public boolean equals(Object o) {
        if (!(o instanceof ID)) return false;
        ID id = (ID) o;
        return val.equals(id.val);
    }
 
    private int compareTo(ID id) {
        return val.compareTo(id.val);
    }
    
    /**
     * Distance from this to id clockwise around identifier circle.
     * If this and id same, then distance is zero.
     *
     * dist in range 0 .. modulus - 1 inclusive.
     *
     */
    
    public BigInteger dist(ID id) {
        if (this.val.compareTo(id.val) <= 0) {  // this <= id
            return id.val.subtract(this.val);
        } else {  // this > id
            return id.val.add(modulus).subtract(this.val);
        }
    }

    /** this in (id1, id2]*  

      See start of file for description of the `star' variation on a range
    */
    public boolean inOpenClosedStarRange(ID id1, ID id2) {
        
        // d(this, id2) < d(id1,id2)

        return this.dist(id2).compareTo(id1.dist(id2)) < 0;
    }
    
    
    /** this in (id1, id2] */
    public boolean inOpenClosedRange(ID id1, ID id2) {
        
        return ! inOpenClosedStarRange(id2, id1);
    }
    
     /** this in (id1,id2) */
    public boolean inOpenOpenRange(ID id1, ID id2) {

        return (!this.equals(id2)) && this.inOpenClosedRange(id1, id2);
    }
    
     /** this in [id1,id2) */
    public boolean inClosedOpenRange(ID id1, ID id2) {

        return this.equals(id1) || this.inOpenOpenRange(id1,id2);
    }

     /** this in [id1,id2)*   */
    public boolean inClosedOpenStarRange(ID id1, ID id2) {

        return  ! inClosedOpenRange(id2, id1);
    }
    
    
    
    /** 
     * @return (this.val + 1) mod 2^size
     */
    
    public ID addOne() {
        return new ID( val.add(BigInteger.ONE).mod(modulus) );        
    }
    
    /** 
     * @return (this.val + 2^i) mod 2^size
     */
    public ID addPowerTwo(int i) {
        return new ID( val.add(powerTwo(i)).mod(modulus) );        
    }
    /** 
     * @return (this.val - 2^i) mod 2^size
     */
    public ID subtractPowerTwo(int i) {
        return new ID( val.subtract(powerTwo(i)).mod(modulus) );        
    }
    
    public String toIntString(int radix) {return val.toString(radix); }

    public String toString() {
        return "ID[" + val + "]";
    }

    
}
