/*
 * CNTest.java
 *
 */

package dhash;

import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.RemoteException;
import java.net.MalformedURLException;
import java.rmi.NotBoundException;

//  Junit 4 
import org.junit.*;  
import static org.junit.Assert.*;
import org.junit.runner.JUnitCore;

import java.io.IOException;
import java.util.logging.*;

import util.AuxProcess;
/**
 *
 * @author  Paul Jackson
 *
 * 
 */

public class CNTest {
    //== Static fields / initializers =========================================
    
    private static Logger logger = Logger.getLogger("dhash.CNTest");
    
    //== Constructor ==========================================================
    /** Create a new instance of CNTest */
    
    public CNTest() {}
    
    //== Main Method ======================================================
    
    /**
     * Initiate JUnit test using text-gui-based reporting.
     *
     *
     */
    public static void main(String[] args)
        throws IOException,
               MalformedURLException,
               NotBoundException  {

        // Store the run mode in the System property list.
        // This approach should allow for passing in mode in cases
        // tests bypass use of main. (e.g. if tests run from IDE such
        // as eclipse).

        if (args.length > 0) {
            System.setProperty("dhash.mode",args[0]);
        }
        if (args.length > 1) {
            System.setProperty("dhash.time",args[1]);
        }
        JUnitCore.main("dhash.CNTest");
    }
    
    //== Utility Methods ==================================================

    private static void delay(int n) {
        try {
            Thread.sleep(n);
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
    }

    //== Setup and teardown for whole test suite ============================

    // Pause time between major operations

    private static int time1;

    // Allow for two auxiliary processes:
    private static Process regProc;    // RMI Registry process
    private static Process domBProc;   // 2nd "Domain B" process
                                       // Main process can double as domainA.
   
    // Factories for CNSlave objects in each JVM.

    private static CNFactory domAFactory;
    private static CNFactory domBFactory;
 
    // Mode of operation
    // A = Automatic multi-process
    // M = Manual multi-process
    // S = Single process

    private static String mode;

    // Setup loggers, startup auxiliary processes, and initialise reference
    // to a CNSlave factory in each domain.

    @BeforeClass
    public static void suiteSetUp()
        throws IOException,
               MalformedURLException,
               NotBoundException  {

        // Root logger is initialised with a ConsoleHandler with 
        // level INFO.
        // Enable handling of all log records by this handler.

        Logger rootLogger = Logger.getLogger("");
        Handler rootHandler = rootLogger.getHandlers()[0];
        rootHandler.setLevel(Level.ALL);

        // Enable logging
        // Logger.getLogger("dhash.ChordNodeImp").setLevel(Level.FINE);
        // Logger.getLogger("dhash.ChordNodeImp").setLevel(Level.FINER);

        // Disable INFO level logging for this class.
        logger.setLevel(Level.WARNING);
        // logger.setLevel(Level.FINEST);

        mode = System.getProperty("dhash.mode", "A");
        String time1Str = System.getProperty("dhash.time", "1000");
        time1 = Integer.parseInt(time1Str);

        // Automatic startup of auxiliary processes.
        if (mode.equals("A")) {
            System.out.println("Starting RMI Registry sub-process");

            String classPath = System.getProperty("java.class.path");
            regProc = AuxProcess.exec("rmiregistry",
                                      "rmiregistry -J-cp -J" + classPath);

            delay(time1); // Allow time for RMI Registry to start listening

            System.out.println("Starting Domain B sub-process");
            domBProc = AuxProcess.exec
                ("CNFactory B",
                 "java -cp " + classPath + " dhash.CNFactoryImp B");
            delay(time1);
        }

        // Get factories for both domA and domB.
        
        domAFactory = CNFactoryImp.getFactory();
        if (mode.equals("S")) {
            domBFactory = domAFactory;
        }
        else {
            Registry registry = LocateRegistry.getRegistry();  
            domBFactory = (CNFactory) 
                registry.lookup("cn-factory/domB");
        }

    }

    @AfterClass
    public static void suiteTearDown() {
        if (mode.equals("A")) {
            // Can get exceptions in subprocess output echoing if
            // don't pause here.
            delay(time1);
            System.out.println("\nKilling sub-processes");
            AuxProcess.shutDown(domBProc);
            AuxProcess.shutDown(regProc);
            System.out.println("Sub-processes dead");
        }
    }

    //== Test Methods =====================================================
    
    /**
     * Basic Test
     *
     * The example given in SIGCOMM 01 paper on Chord protocol..
     *
     * Set size = 3.  Create 4 nodes.  Insert into Ring. Do lookups
     * for each.  Check get correct nodes.
     *
     *
     */
    @Test
    public void test1() throws RemoteException {
        
        logger.info("Starting test1 method");
        
        domAFactory.setIDSize(3);
        domBFactory.setIDSize(3);
        
        // Could refine makeCNSlave implementations to augment name
        // of ChordNode with name of domain.  Leave for now.
        
        CNSlave n0 = domAFactory.makeCNSlave("node0", new ID("0"));
        CNSlave n1 = domAFactory.makeCNSlave("node1", new ID("1"));
        CNSlave n3 = domBFactory.makeCNSlave("node3", new ID("3"));
        CNSlave n6 = domBFactory.makeCNSlave("node6", new ID("6"));
        
        logger.info("4 nodes constructed");
        
        n0.initForNewRing();
        n1.join(n0);
        
        logger.info("Formed ring of 2 nodes:");
        logger.info(n0.toLongString());
        logger.info(n1.toLongString());
                
        n3.join(n1);
        
        logger.info("Formed ring of 3 nodes:");
        logger.info(n0.toLongString());
        logger.info(n1.toLongString());
        logger.info(n3.toLongString());
        
        // Check a few cases.
        
        assertEquals(n1.getID(), n0.findSuccessor(new ID("1")).getID());
        assertEquals(n1.getID(), n1.findSuccessor(new ID("1")).getID());
        assertEquals(n1.getID(), n3.findSuccessor(new ID("1")).getID());
        assertEquals(n3.getID(), n1.findSuccessor(new ID("2")).getID());
        assertEquals(n3.getID(), n1.findSuccessor(new ID("3")).getID());
        assertEquals(n0.getID(), n1.findSuccessor(new ID("4")).getID());
        assertEquals(n0.getID(), n1.findSuccessor(new ID("5")).getID());
        assertEquals(n0.getID(), n1.findSuccessor(new ID("6")).getID());
        assertEquals(n0.getID(), n1.findSuccessor(new ID("7")).getID());
        assertEquals(n0.getID(), n1.findSuccessor(new ID("0")).getID());
        
               
        n6.join(n1);
        
        logger.info("Node 6 joins");
        logger.info(n0.toLongString());
        logger.info(n1.toLongString());
        logger.info(n3.toLongString());
        logger.info(n6.toLongString());
        
         
    }
    
}
