/*
 * DHNTest.java
 *
 * JUnit 4 test of DHashNode distributed hash table
 *
 * @author  Paul Jackson
 */

package dhash;

//  Junit 4 
import org.junit.*;  
import static org.junit.Assert.*;
import org.junit.runner.JUnitCore;

import java.util.logging.*;

import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.RemoteException;
import java.net.MalformedURLException;
import java.rmi.AlreadyBoundException;
import java.rmi.NotBoundException;
import java.rmi.server.UnicastRemoteObject;

import java.io.IOException;

import util.AuxProcess;

//==EXTRA IMPORTS FOR SOLUTION ===============================================
//Keep this banner





public class DHNTest {
    //== Static fields / initializers =========================================
    
    private static Logger logger = Logger.getLogger("dhash.DHNTest");
     
    //== Constructor ========================================================= 
    /** Create a new instance of DHNTest */

    public DHNTest() {}
    
    //== Main Methods ========================================================

    /**
     * Initiate JUnit test using text-gui-based reporting.
     * Options:
     * 1st arg:
     *   S = Run sequential single-JVM test.
     *   M = Run manually initiated multi-JVM test.
     *   A = Run automatically initiated multi-JVM test. (Default)
     * 2nd arg:
     *   Time for pauses.  (Default 1000ms)
     */

    public static void main(String[] args)
        throws IOException,
               MalformedURLException,
               NotBoundException  {

        if (args.length > 0) {
            System.setProperty("dhash.mode",args[0]);
        }
        if (args.length > 1) {
            System.setProperty("dhash.time",args[1]);
        }
        JUnitCore.main("dhash.DHNTest");
    }

    //== Utility Methods ==================================================

    private static void delay(int n) {
        try {
            Thread.sleep(n);
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
    }

    //== Setup and teardown for whole test suite ===========================

    

    private static String mode;   // Mode of operation
    private static int time1;     // Pause time between major operations

    // Handles for auxiliary sub processes. 
    // Multi Java Runtime test configured so only main JUnit code runs in 
    // primary process.

    private static Process regProc;   // RMI Registry
    private static Process domAProc;  // Domain A process
    private static Process domBProc;  // Domain B process

    private static DHNFactory domAFactory;
    private static DHNFactory domBFactory;

    private static MonitorImp monitor;

    // Setup common to all tests

    // All this code put here rather than in main method, because main()
    // is bypassed when JUnit tests are run directly from some IDEs.

    //-- suiteSetUp -------------------------------------------------------

    @BeforeClass
    public static void suiteSetUp()
        throws IOException,                                   
               MalformedURLException,
               NotBoundException {
        
        mode = System.getProperty("dhash.mode", "A");
        String time1Str = System.getProperty("dhash.time", "1000");
        time1 = Integer.parseInt(time1Str);

        // Initialise monitor

        monitor = new MonitorImp();

        // Configure logging

        // Enable handling of all levels by the ConsoleHandler
        // associated with the root logger. (Handles INFO and above by default)

        Logger.getLogger("").getHandlers()[0].setLevel(Level.ALL);

        if (mode.equals("S")) {

            // Create and install new handler for custom tracing of
            // ChordNode methods.

            // Only need to do this for S mode, since in M and A modes
            // handling is only used in sub-processes.
            
            Handler cNIHandler = new CNHandler(monitor);
            cNIHandler.setLevel(Level.ALL);

            Logger cNILogger = Logger.getLogger("dhash.ChordNodeImp");
            cNILogger.addHandler(cNIHandler);
            cNILogger.setUseParentHandlers(false);
        }

        String classPath = System.getProperty("java.class.path");
        
        // Start a Java Registry sub-process.

        if (mode.equals("A")) {
            System.out.println("Starting RMI Registry sub-process");
            regProc = AuxProcess.exec("rmiregistry",
                                      "rmiregistry -J-cp -J" + classPath);
            delay(time1); // Allow time for RMI Registry to start listening
        }
        
        // Register the monitor object with the Registry.

        if (mode.equals("A") || mode.equals("M")) {

            try {
                // The 0 as 2nd arg of exportObject specifies an
                // anonymous port: Port number will be chosen at
                // runtime.

                Monitor monitorStub = (Monitor)
                    UnicastRemoteObject.exportObject(monitor, 0);

                Registry registry = LocateRegistry.getRegistry();  

                registry.bind("dhash/monitor", monitorStub);

            } catch (Exception e) {
                System.out.println("DHNTest monitor export exception: "
                                   + e.getMessage());
                e.printStackTrace();
            }
        }
        
        // Start the subprocesses for the DHNSlaveImp factories

        // On startup, these subprocesses notify the monitor of their
        // existence.

        if (mode.equals("A")) {
            System.out.println("Starting Domain A sub-process");
            domAProc = AuxProcess.exec
                ("DHNFactory A",
                 "java -cp " + classPath + " dhash.DHNFactoryImp A");
            System.out.println("Starting Domain B sub-process");
            domBProc = AuxProcess.exec
                ("DHNFactory B",
                 "java -cp " + classPath + " dhash.DHNFactoryImp B");
            delay(time1);
        }

        // Initialise references to the DHNSlaveImp factories

        if (mode.equals("S")) {
            domAFactory = DHNFactoryImp.getFactory();
            domBFactory = domAFactory;
        }
        else { // cmd = "M" or "A"

            // Wait for slave factories have recorded references to
            // themselves with the monitor.

            while (monitor.mapSize() < 2) {
                System.out.println("DHNTest: waiting for slaves to start");
                delay(time1);
                delay(time1);
            }

            domAFactory = (DHNFactory) monitor.get("A");
            domBFactory = (DHNFactory) monitor.get("B");

        }
    }

    //--suiteTearDown() ---------------------------------------------------
    
    @AfterClass
    public static void suiteTearDown() {

        if (mode.equals("A")) {
            // Can get exceptions in subprocess output echoing if
            // don't pause here.
            delay(time1);
            System.out.println("\nKilling sub-processes");
            AuxProcess.shutDown(domAProc);
            AuxProcess.shutDown(domBProc);
            AuxProcess.shutDown(regProc);
            System.out.println("Sub-processes dead");
        }
    }

    //== Instance fields initialised at start of each test ================

    String ka; 
    String kb;
    String kc;
    String kd;
    String ke;
        
    String va;
    String vb;
    String vc;
    String vd;
    String ve;
        

    //== Setup and teardown for each test ==================================

    // Setup common to all tests

    @Before
    public void setUp() {
         
        ka = "ka";
        kb = "kb";
        kc = "kc";
        kd = "kd";
        ke = "ke";
        
        va = "va";
        vb = "vb";
        vc = "vc";
        vd = "vd";
        ve = "ve";
    }        

    @After
    public void tearDown() {

        // Try to ensure all logger output appears before JUnit output.
        logger.getParent().getHandlers()[0].flush();
        System.err.flush();
    }

    //== Test Methods ========================================================


    /**
     * Do basic checks of DHash table operation
     * 
     * Test values chosen so that key value pairs redistributed on new
     * node joins
     */

    //== SOLUTION CODE =======================================================
    //Keep this banner

    @Test
    public void test1() throws RemoteException {

        ID.setSize(4);
        domAFactory.setIDSize(4);
        domBFactory.setIDSize(4);

        DHNSlave n1 = domAFactory.makeDHNSlave("node1", new ID("1"));
        DHNSlave n7 = domBFactory.makeDHNSlave("node7", new ID("7"));
        DHNSlave n9 = domBFactory.makeDHNSlave("node9", new ID("9"));
        DHNSlave n15 = domAFactory.makeDHNSlave("node15", new ID("15"));

        n7.initForNewRing();
        
        n7.put(ka,va);
        n7.put(kb,vb);
        n7.put(kc,vc);

        n9.join(n7);
        n1.join(n9);
        n15.join(n1);
        
        assertTrue(n1.contains(kb));
        assertFalse(n9.contains(kd));

        n7.put(kd,vd);
        n7.put(ke,ve);
        
        assertEquals(va, n1.get(ka) );
        assertEquals(va, n7.get(ka) );
        assertEquals(vb, n1.get(kb) );
        assertEquals(vc, n1.get(kc) );
        assertEquals(vd, n1.get(kd) );

        // Enable tracing of findSuccessor().(Uncomment for Part 1, Task B)
        // domAFactory.setLoggingLevel("dhash.ChordNodeImp", Level.FINE);
        // domBFactory.setLoggingLevel("dhash.ChordNodeImp", Level.FINE);

        n1.getDHashNode().getChordNode().findSuccessor(new ID("5"));
        System.out.println();
        n9.getDHashNode().getChordNode().findSuccessor(new ID("8"));

        // Disable tracing of findSuccessor(). (Uncomment for Part 1, Task B)
        // domAFactory.setLoggingLevel("dhash.ChordNodeImp", Level.INFO);
        // domBFactory.setLoggingLevel("dhash.ChordNodeImp", Level.INFO);

        n1.remove(kd);

        assertTrue(n1.contains(ka));
        assertTrue(n9.contains(kb));
        assertTrue(n7.contains(kc));
        assertTrue(n1.contains(ke));
        assertFalse(n9.contains(kd));
        
        
    }

    //== END SOLUTION CODE ====================================================
    //Keep this banner

    /**
     * 
     * Demonstrate use of DHAppNode interface.
     *
     * Use same configuration as with Test 1.
     *
     * ****** PLEASE DELETE THIS TEST FOR THE PART 1 SUBMISSION *************
     */
    @Test
    public void test2() throws RemoteException {
        
        ID.setSize(4);
        domAFactory.setIDSize(4);
        domBFactory.setIDSize(4);

        DHNSlave n1 = domAFactory.makeDHNSlave("node1", new ID("1"));
        DHNSlave n7 = domBFactory.makeDHNSlave("node7", new ID("7"));
        DHNSlave n9 = domBFactory.makeDHNSlave("node9", new ID("9"));
        DHNSlave n15 = domAFactory.makeDHNSlave("node15", new ID("15"));

        
        n7.initForNewRing();
        
        n7.put(ka,va);
        n7.put(kb,vb);
        n7.put(kc,vc);

        n9.join(n7);
        n1.join(n9);
        n15.join(n1);

        // Test two stage value retrieval via SimpleDHApp object.

        // The dHAppGet(ka) call stores a reference to the value for key ka on 
        // the DHash application node paired with the DHash node holding
        // the ka value pair.
        // dhan is then a remote reference to this DHash Application node.

        DHAppNode dhan = n1.dHAppGet(ka);

        // The DHNSlaveImp class uses the SimpleDHAppImp class as the
        // DHash application node.  Remote references to DHNSlaveImp objects
        // have interface type SimpleDHApp.  Hence we can cast dhan to
        // the SimpleDHApp type.

        SimpleDHApp sdha = (SimpleDHApp) dhan;

        // Go fetch the value from the remote SimpleDHAppImp class.
        // The getValue() call on the SimpleDHAppImp object serialises
        // the value and the getValue() call on the local stub of
        // type SimpleDHApp deserialises the object.
        
        Object fetchedVa = sdha.getValue();

        // We know the value is actually a String, so cast it down.

        String fetchedVaString = (String) fetchedVa;

        // Check this fetchedVaString.

        assertEquals(va, fetchedVaString);
        
        
    }


}
